@@ -4,6 +4,12 @@ use utf8;
 use 5.016;
 use warnings;
 
+use Memoize;
+
+memoize('_get_video_info');
+memoize('_extract_from_ytdl');
+memoize('_extract_from_invidious');
+
 use parent qw(
   WWW::YoutubeViewer::Search
   WWW::YoutubeViewer::Videos
@@ -24,7 +30,7 @@ WWW::YoutubeViewer - A very easy interface to YouTube.
 
 =cut
 
-our $VERSION = '3.7.8';
+our $VERSION = '3.7.9';
 
 =head1 SYNOPSIS
 
@@ -41,31 +47,31 @@ my %valid_options = (
 
     # Main options
     v               => {valid => q[],                                                    default => 3},
-    page            => {valid => [qr/^(?!0+\z)\d+\z/],                                   default => 1},
-    http_proxy      => {valid => [qr{.}],                                                default => undef},
-    hl              => {valid => [qr/^\w+(?:[\-_]\w+)?\z/],                              default => undef},
+    page            => {valid => qr/^(?!0+\z)\d+\z/,                                     default => 1},
+    http_proxy      => {valid => qr/./,                                                  default => undef},
+    hl              => {valid => qr/^\w+(?:[\-_]\w+)?\z/,                                default => undef},
     maxResults      => {valid => [1 .. 50],                                              default => 10},
-    topicId         => {valid => [qr/^./],                                               default => undef},
+    topicId         => {valid => qr/./,                                                  default => undef},
     order           => {valid => [qw(relevance date rating viewCount title videoCount)], default => undef},
-    publishedAfter  => {valid => [qr/^\d+/],                                             default => undef},
-    publishedBefore => {valid => [qr/^\d+/],                                             default => undef},
-    channelId       => {valid => [qr/^[-\w]{2,}\z/],                                     default => undef},
+    publishedAfter  => {valid => qr/^\d+/,                                               default => undef},
+    publishedBefore => {valid => qr/^\d+/,                                               default => undef},
+    channelId       => {valid => qr/^[-\w]{2,}\z/,                                       default => undef},
     channelType     => {valid => [qw(any show)],                                         default => undef},
 
     # Video only options
     videoCaption    => {valid => [qw(any closedCaption none)],     default => undef},
     videoDefinition => {valid => [qw(any high standard)],          default => undef},
-    videoCategoryId => {valid => [qr/^\d+\z/],                     default => undef},
+    videoCategoryId => {valid => qr/^\d+\z/,                       default => undef},
     videoDimension  => {valid => [qw(any 2d 3d)],                  default => undef},
     videoDuration   => {valid => [qw(any short medium long)],      default => undef},
     videoEmbeddable => {valid => [qw(any true)],                   default => undef},
     videoLicense    => {valid => [qw(any creativeCommon youtube)], default => undef},
     videoSyndicated => {valid => [qw(any true)],                   default => undef},
     eventType       => {valid => [qw(completed live upcoming)],    default => undef},
-    chart           => {valid => [qw(mostPopular)],                default => 'mostPopular'},
+    chart           => {valid => [qw(mostPopular)],                default => undef},
 
-    regionCode        => {valid => [qr/^[A-Z]{2}\z/i],         default => undef},
-    relevanceLanguage => {valid => [qr/^[a-z](?:\-\w+)?\z/i],  default => undef},
+    regionCode        => {valid => qr/^[A-Z]{2}\z/i,           default => undef},
+    relevanceLanguage => {valid => qr/^[a-z]+(?:\-\w+)?\z/i,   default => undef},
     safeSearch        => {valid => [qw(none moderate strict)], default => undef},
     videoType         => {valid => [qw(any episode movie)],    default => undef},
 
@@ -73,26 +79,27 @@ my %valid_options = (
     subscriptions_order => {valid => [qw(alphabetical relevance unread)], default => undef},
 
     # Misc
-    debug       => {valid => [0 .. 3],     default => 0},
-    lwp_timeout => {valid => [qr/^\d+\z/], default => 1},
-    config_dir  => {valid => [qr/^./],     default => q{.}},
-    cache_dir   => {valid => [qr/^./],     default => q{.}},
+    debug       => {valid => [0 .. 3],   default => 0},
+    timeout     => {valid => qr/^\d+\z/, default => 10},
+    config_dir  => {valid => qr/^./,     default => q{.}},
+    cache_dir   => {valid => qr/^./,     default => q{.}},
+    cookie_file => {valid => qr/^./,     default => undef},
 
     # Booleans
-    lwp_env_proxy => {valid => [1, 0], default => 1},
-    escape_utf8   => {valid => [1, 0], default => 0},
-    prefer_mp4    => {valid => [1, 0], default => 0},
-    prefer_av1    => {valid => [1, 0], default => 0},
+    env_proxy   => {valid => [1, 0], default => 1},
+    escape_utf8 => {valid => [1, 0], default => 0},
+    prefer_mp4  => {valid => [1, 0], default => 0},
+    prefer_av1  => {valid => [1, 0], default => 0},
 
     # API/OAuth
-    key           => {valid => [qr/^.{15}/], default => undef},
-    client_id     => {valid => [qr/^.{15}/], default => undef},
-    client_secret => {valid => [qr/^.{15}/], default => undef},
-    redirect_uri  => {valid => [qr/^.{15}/], default => 'urn:ietf:wg:oauth:2.0:oob'},
-    access_token  => {valid => [qr/^.{15}/], default => undef},
-    refresh_token => {valid => [qr/^.{15}/], default => undef},
+    key           => {valid => qr/^.{15}/, default => undef},
+    client_id     => {valid => qr/^.{15}/, default => undef},
+    client_secret => {valid => qr/^.{15}/, default => undef},
+    redirect_uri  => {valid => qr/^.{15}/, default => 'urn:ietf:wg:oauth:2.0:oob'},
+    access_token  => {valid => qr/^.{15}/, default => undef},
+    refresh_token => {valid => qr/^.{15}/, default => undef},
 
-    authentication_file => {valid => [qr/^./], default => undef},
+    authentication_file => {valid => qr/^./, default => undef},
 
     # No input value allowed
     feeds_url        => {valid => q[], default => 'https://www.googleapis.com/youtube/v3/'},
@@ -101,8 +108,10 @@ my %valid_options = (
     video_info_args  => {valid => q[], default => '?video_id=%s&el=detailpage&ps=default&eurl=&gl=US&hl=en'},
     www_content_type => {valid => q[], default => 'application/x-www-form-urlencoded'},
 
+#<<<
     # LWP user agent
-    lwp_agent => {valid => [qr/^.{5}/], default => 'Mozilla/5.0 (X11; U; Linux i686; gzip; en-US) Chrome/10.0.648.45'},
+    user_agent => {valid => qr/^.{5}/, default => 'Mozilla/5.0 (Windows NT 10.0; Win64; gzip; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.0.0 Safari/537.36'},
+#>>>
 );
 
 sub _our_smartmatch {
@@ -110,7 +119,7 @@ sub _our_smartmatch {
 
     $value // return 0;
 
-    if (ref($arg) eq '') {
+    if (not ref($arg)) {
         return ($value eq $arg);
     }
 
@@ -132,7 +141,7 @@ sub _our_smartmatch {
 
     foreach my $key (keys %valid_options) {
 
-        if (ref $valid_options{$key}{valid} eq 'ARRAY') {
+        if (ref($valid_options{$key}{valid})) {
 
             # Create the 'set_*' subroutines
             *{__PACKAGE__ . '::set_' . $key} = sub {
@@ -235,9 +244,12 @@ sub set_lwp_useragent {
 
     $self->{lwp} = $lwp->new(
 
-        timeout       => $self->get_lwp_timeout,
+        cookie_jar    => {},                      # temporary cookies
+        timeout       => $self->get_timeout,
         show_progress => $self->get_debug,
-        agent         => $self->get_lwp_agent,
+        agent         => $self->get_user_agent,
+
+        ssl_opts => {verify_hostname => 1},
 
         $lwp eq 'LWP::UserAgent::Cached'
         ? (
@@ -264,7 +276,7 @@ sub set_lwp_useragent {
           )
         : (),
 
-        env_proxy => (defined($self->get_http_proxy) ? 0 : $self->get_lwp_env_proxy),
+        env_proxy => (defined($self->get_http_proxy) ? 0 : $self->get_env_proxy),
     );
 
     require LWP::ConnCache;
@@ -277,16 +289,41 @@ sub set_lwp_useragent {
     };
 
     my $agent = $self->{lwp};
-    $agent->ssl_opts(Timeout => 30);
+    $agent->ssl_opts(Timeout => $self->get_timeout);
     $agent->default_header('Accept-Encoding' => $accepted_encodings);
     $agent->conn_cache($cache);
     $agent->proxy(['http', 'https'], $self->get_http_proxy) if defined($self->get_http_proxy);
 
-    #my $http_proxy = $agent->proxy('http');
-    #if (defined($http_proxy)) {
-    #    $agent->proxy('https', $http_proxy) if (!defined($agent->proxy('https')));
-    #    $agent->timeout(30);
-    #}
+    my $cookie_file = $self->get_cookie_file;
+
+    if (defined($cookie_file) and -f $cookie_file) {
+
+        if ($self->get_debug) {
+            say STDERR ":: Using cookies from: $cookie_file";
+        }
+
+        ## Netscape HTTP Cookies
+
+        # Chrome extension:
+        #   https://chrome.google.com/webstore/detail/cookiestxt/njabckikapfpffapmjgojcnbfjonfjfg
+
+        # Firefox extension:
+        #   https://addons.mozilla.org/en-US/firefox/addon/cookies-txt/
+
+        # See also:
+        #   https://github.com/ytdl-org/youtube-dl#how-do-i-pass-cookies-to-youtube-dl
+
+        require HTTP::Cookies::Netscape;
+
+        my $cookies = HTTP::Cookies::Netscape->new(
+                                                   hide_cookie2 => 1,
+                                                   autosave     => 1,
+                                                   file         => $cookie_file,
+                                                  );
+
+        $cookies->load;
+        $agent->cookie_jar($cookies);
+    }
 
     push @{$self->{lwp}->requests_redirectable}, 'POST';
     return $self->{lwp};
@@ -343,6 +380,10 @@ sub lwp_get {
     $url // return;
     $self->{lwp} // $self->set_lwp_useragent();
 
+    if (not defined($self->get_key)) {
+        return undef if not $opt{simple};
+    }
+
     my %lwp_header = ($opt{simple} ? () : $self->_auth_lwp_header);
     my $response   = $self->{lwp}->get($url, %lwp_header);
 
@@ -489,14 +530,92 @@ sub _make_feed_url {
     $self->get_feeds_url() . $path . '?' . $self->default_arguments(%args);
 }
 
+sub get_invidious_instances {
+    my ($self) = @_;
+
+    require File::Spec;
+    my $instances_file = File::Spec->catfile($self->get_config_dir, 'instances.json');
+
+    # Get the "instances.json" file when the local copy is too old or non-existent
+    if ((not -e $instances_file) or (-M _) > 1 / 24) {
+
+        require LWP::UserAgent;
+
+        my $lwp = LWP::UserAgent->new(timeout => $self->get_timeout);
+        $lwp->show_progress(1) if $self->get_debug;
+        my $resp = $lwp->get("https://instances.invidio.us/instances.json");
+
+        $resp->is_success() or return;
+
+        my $json = $resp->decoded_content() || return;
+        open(my $fh, '>', $instances_file) or return;
+        print $fh $json;
+        close $fh;
+    }
+
+    open(my $fh, '<', $instances_file) or return;
+
+    my $json_string = do {
+        local $/;
+        <$fh>;
+    };
+
+    $self->parse_json_string($json_string);
+}
+
+sub select_good_invidious_instances {
+    my ($self) = @_;
+
+    state $instances = $self->get_invidious_instances;
+
+    ref($instances) eq 'ARRAY' or return;
+
+    my %ignored = (
+                   'yewtu.be'                 => 1,
+                   'invidiou.site'            => 0,
+                   'invidious.xyz'            => 1,
+                   'vid.mint.lgbt'            => 1,
+                   'invidious.ggc-project.de' => 1,
+                   'invidious.toot.koeln'     => 1,
+                   'invidious.snopyta.org'    => 1,    # too popular == too slow
+                  );
+
+    my @candidates =
+      grep { not $ignored{$_->[0]} }
+      grep { ref($_->[1]{monitor}) eq 'HASH' ? ($_->[1]{monitor}{statusClass} eq 'success') : 1 }
+      grep { lc($_->[1]{type} // '') eq 'https' } @$instances;
+
+    if ($self->get_debug) {
+        print STDERR ":: Found ", scalar(@candidates), " invidious instances.\n";
+    }
+
+    return @candidates;
+}
+
 sub _extract_from_invidious {
     my ($self, $videoID) = @_;
 
-    my @instances = qw(
-      invidious.snopyta.org
-      invidious.13ad.de
-      invidious.tube
-      );
+    my @instances = $self->select_good_invidious_instances();
+
+    if (@instances) {
+        require List::Util;
+        @instances = List::Util::shuffle(map { $_->[0] } @instances);
+        push @instances, 'invidious.snopyta.org';
+        push @instances, 'invidious.13ad.de';
+    }
+    else {
+        @instances = qw(
+          invidious.tube
+          invidious.site
+          invidious.fdn.fr
+          invidious.snopyta.org
+          invidious.13ad.de
+          );
+    }
+
+    if ($self->get_debug) {
+        print STDERR ":: Invidious instances: @instances\n";
+    }
 
     my $tries      = 2 * scalar(@instances);
     my $instance   = shift(@instances);
@@ -505,7 +624,7 @@ sub _extract_from_invidious {
 
     my $resp = $self->{lwp}->get($url);
 
-    while (not $resp->is_success() and $resp->status_line() =~ /read timeout/i and --$tries >= 0) {
+    while (not $resp->is_success() and --$tries >= 0) {
         $url  = sprintf($url_format, shift(@instances), $videoID) if (@instances and ($tries % 2 == 0));
         $resp = $self->{lwp}->get($url);
     }
@@ -538,8 +657,15 @@ sub _extract_from_ytdl {
 
     $self->_ytdl_is_available() || return;
 
-    my $json = $self->proxy_stdout('youtube-dl', '--all-formats', '--dump-single-json',
-                                   quotemeta("https://www.youtube.com/watch?v=" . $videoID));
+    my @ytdl_cmd = ('youtube-dl', '--all-formats', '--dump-single-json');
+
+    my $cookie_file = $self->get_cookie_file;
+
+    if (defined($cookie_file) and -f $cookie_file) {
+        push @ytdl_cmd, '--cookies', $cookie_file;
+    }
+
+    my $json = $self->proxy_stdout(@ytdl_cmd, quotemeta("https://www.youtube.com/watch?v=" . $videoID));
 
     my $ref = $self->parse_json_string($json);
 
@@ -838,6 +964,11 @@ sub get_streaming_urls {
         }
     }
 
+    if ($self->get_debug) {
+        my $count = scalar(@streaming_urls);
+        say STDERR ":: Found $count streaming URLs...";
+    }
+
     # Try again with youtube-dl
     if (!@streaming_urls or $info{status} =~ /fail|error/i) {
         @streaming_urls = $self->_fallback_extract_urls($videoID);
